/*=========================================================================*/
/**
	@file		wpgetimg.c
	@brief		wizpy image get utility
	@version	0.02
	@Date
	@author
/*
/*=========================================================================*/

#include	<stdio.h>

#define	SCREEN_WIDTH	128				// wizpy screen width & height
#define	SCREEN_HEIGHT	160
#define	SCREEN_BPP		32

										// bit15:0 bit14-10:R bit9-5:G bit4-0:B
#define	WIZPY_MASK_R	0x7C00			// 0b0111_1100_0000_0000
#define	WIZPY_MASK_G	0x03E0			// 0b0000_0011_1110_0000
#define	WIZPY_MASK_B	0x001F			// 0b0000_0000_0001_1111

/* BitmapFileHeader */
typedef struct BITMAPFILEHEADER_tag {
	unsigned short bfType;
	unsigned int bfSize;
	unsigned short bfReserved1;
	unsigned short bfReserved2;
	unsigned int bfOffBits;
} BitmapFileHeader;

/* BitmapInfoHeader */
typedef struct BITMAPINFOHEADER_tag {
	unsigned int biSize;
	int biWidth;
	int biHeight;
	unsigned short biPlanes;
	unsigned short biBitCount;
	unsigned int biCompression;
	unsigned int biSizeImage;
	int biXPixPerMeter;
	int biYPixPerMeter;
	unsigned int biClrUsed;
	unsigned int biClrImportant;
} BitmapInfoHeader;

BitmapFileHeader BmpFileHeader;
BitmapInfoHeader BmpInfoHeader;

unsigned int address = 0x00;			// wizpy image data address
unsigned int width = 128;				// image width
unsigned int height = 160;				// image height
unsigned int bpp = 24;					// image color bpp

										// wizpy image data buffer
unsigned char wizbuf[ SCREEN_WIDTH * SCREEN_HEIGHT * 2 ];
										// BMP(32bpp) image data buffer
unsigned char bmpbuf[ SCREEN_WIDTH * SCREEN_HEIGHT * 4 ];


/**
	@fn			int readWizData( FILE *fp )
	@brief		read data to image buffer
	@param		*fp : wizpy firmware file pointer
	@return		>0 : read size
	@return		-1 : read failed
 */
int readWizData( FILE *fp )
{
	int rsize;

	/* seek input file */
	if( fseek( fp, address, SEEK_SET ) != 0 ) {
		printf( "input file seek failed.\n" );
		fclose( fp );
		return 1;
    }

	/* read input file */
	rsize = fread( wizbuf, ( width * height * 2 ), 1, fp );
	if( rsize < 1 ) {
		printf( "input file read failed.\n" );
		fclose( fp );
		return 1;
	}

	return rsize;
}


/**
	@fn			void writeBMP( const char *filename )
	@brief		write BMP file
	@param		*filename : output bitmap image filename
	@return		null
 */
void writeBMP( const char *filename )
{
	FILE *fwp = NULL;
	int adjust = ( width % 4 );			// adjust bytes (4-byte boundary for each scan line)
	int sizeimage = 0;					// size of data part

	/* set BitmapFileHeader */
	BmpFileHeader.bfType = 0x4D42;		/* "BM" */
	if( bpp == 32 ) {
		sizeimage = ( width * height * 4 );
	}
	else if( bpp == 24 ) {
		sizeimage = ( height * ( width * 3 + adjust ) );
	}
	else {
		printf( "image size error\n" );
		return;
	}
	BmpFileHeader.bfSize = 14 + 40 + sizeimage;
	BmpFileHeader.bfReserved1 = 0;
	BmpFileHeader.bfReserved2 = 0;
	BmpFileHeader.bfOffBits = 54;

	/* set BitmapInfoHeader */
	BmpInfoHeader.biSize = 40;
	BmpInfoHeader.biWidth = width;
	BmpInfoHeader.biHeight = height;
	BmpInfoHeader.biPlanes = 1;
	BmpInfoHeader.biBitCount = bpp;
	BmpInfoHeader.biCompression = 0;
	BmpInfoHeader.biSizeImage = sizeimage;
	BmpInfoHeader.biXPixPerMeter = 0;
	BmpInfoHeader.biYPixPerMeter = 0;
	BmpInfoHeader.biClrUsed = 0;
	BmpInfoHeader.biClrImportant = 0;

	/* open output file */
	fwp = fopen( filename, "wb" );
	if( fwp == NULL ) {
		printf( "output file open failed.\n" );
	}

	/* write BitmapFileHeader */
	fwrite( &BmpFileHeader.bfType, sizeof( BmpFileHeader.bfType ), 1, fwp );
	fwrite( &BmpFileHeader.bfSize, sizeof( BmpFileHeader.bfSize ), 1, fwp );
	fwrite( &BmpFileHeader.bfReserved1, sizeof( BmpFileHeader.bfReserved1 ), 1, fwp );
	fwrite( &BmpFileHeader.bfReserved2, sizeof( BmpFileHeader.bfReserved2 ), 1, fwp );
	fwrite( &BmpFileHeader.bfOffBits, sizeof( BmpFileHeader.bfOffBits ), 1, fwp );

	/* write BitmapInfoHeader */
	fwrite( &BmpInfoHeader.biSize, sizeof( BmpInfoHeader.biSize ), 1, fwp );
	fwrite( &BmpInfoHeader.biWidth, sizeof( BmpInfoHeader.biWidth ), 1, fwp );
	fwrite( &BmpInfoHeader.biHeight, sizeof( BmpInfoHeader.biHeight ), 1, fwp );
	fwrite( &BmpInfoHeader.biPlanes, sizeof( BmpInfoHeader.biPlanes ), 1, fwp );
	fwrite( &BmpInfoHeader.biBitCount, sizeof( BmpInfoHeader.biBitCount ), 1, fwp );
	fwrite( &BmpInfoHeader.biCompression, sizeof( BmpInfoHeader.biCompression ), 1, fwp );
	fwrite( &BmpInfoHeader.biSizeImage, sizeof( BmpInfoHeader.biSizeImage ), 1, fwp );
	fwrite( &BmpInfoHeader.biXPixPerMeter, sizeof( BmpInfoHeader.biXPixPerMeter ), 1, fwp );
	fwrite( &BmpInfoHeader.biYPixPerMeter, sizeof( BmpInfoHeader.biYPixPerMeter ), 1, fwp );
	fwrite( &BmpInfoHeader.biClrUsed, sizeof( BmpInfoHeader.biClrUsed ), 1, fwp );
	fwrite( &BmpInfoHeader.biClrImportant, sizeof( BmpInfoHeader.biClrImportant ), 1, fwp );

	/* write bitmap data */
	fwrite( bmpbuf, sizeimage, 1, fwp );

	return;
}


/**
	@fn			void convert24( void )
	@brief		convert wizpy image to 24bpp
	@param		null
	@return		null
 */
void convert24( void )
{
	int x = 0;
	int y = 0;
	int rp = 0;
	int wp = 0;
	unsigned short color16 = 0;
	int adjust = ( width % 4 );			// adjust bytes (4-byte boundary for each scan line)

	for( y = ( height - 1 ); y >= 0; y-- ) {
		for( x = 0; x < width; x++ ) {
			/* read and write pointer */
			rp = ( x + ( y * width ) ) * 2;
			wp = ( x * 3 ) + ( ( height - y - 1 ) * ( width * 3 + adjust ) );
			/* get 16bit color value */
			color16 = ( ( wizbuf[ rp + 1 ] << 8 ) + wizbuf[ rp ] );
			/* convert to 24bit color */
			bmpbuf[ wp ] =  ( ( color16 & WIZPY_MASK_B ) << 3 );
			bmpbuf[ wp + 1 ] = ( ( color16 & WIZPY_MASK_G ) >>  5 ) << 3;
			bmpbuf[ wp + 2 ] = ( ( color16 & WIZPY_MASK_R ) >> 10 ) << 3;
		}
	}

	return;
}


/**
	@fn			void convert32( void )
	@brief		convert wizpy image to 32bpp
	@param		null
	@return		null
 */
void convert32( void )
{
	int x = 0;
	int y = 0;
	int rp = 0;
	int wp = 0;
	unsigned short color16 = 0;

	for( y = 0; y < height; y++ ) {
		for( x = 0; x < width; x++ ) {
			/* read and write pointer */
			rp = ( x + ( y * width ) ) * 2;
			wp = ( x + ( ( height - y - 1 ) * width ) ) * 4;
			/* get 16bit color value */
			color16 = ( ( wizbuf[ rp + 1 ] << 8 ) + wizbuf[ rp ] );
			/* convert to 32bit color */
			bmpbuf[ wp ] =  ( ( color16 & WIZPY_MASK_B ) << 3 );
			bmpbuf[ wp + 1 ] = ( ( color16 & WIZPY_MASK_G ) >>  5 ) << 3;
			bmpbuf[ wp + 2 ] = ( ( color16 & WIZPY_MASK_R ) >> 10 ) << 3;
			bmpbuf[ wp + 3 ] = 0x00;
		}
	}

	return;
}


/**
	@fn			void printUsage( void )
	@brief		print usage
	@param		null
	@return		null
 */
void printUsage( void )
{
	printf( "wizpy firmware image get utility  ver 0.2\n" );
	printf( "Syntax: wpgetimg <firmware-file> <address(hex)> <width> <height> <output-file(.bmp)> [-b bpp]\n" );
	printf( "        or\n" );
	printf( "        wpgetimg <firmware-file> -l <listfile>\n" );
	printf( "\n" );

	return;
}


/**
	@fn			int main( int argc, char *argv[] )
	@brief		main routine
	@param		argc
	@param		*argv[]
	@return		
 */
int main( int argc, char *argv[] )
{
	FILE *fpfirm;						// firmware file
	FILE *fplist;						// list file
	int mode = 0;
	int i = 0;
	char listarg[ 256 ];
	char namearg[ 128 ];
	char filename[ 128 + 4 ];
	unsigned int size;
	int b = 0;


	/* check options */
	if( argc < 2 ) {
		printUsage();
		return 1;
	}
	/* check batch mode */
	else if( argv[2][0] == '-' ) {
		if( argv[2][1] == 'l' ) {
			mode = 1;
		}
		else {
			printf( "unknown option :%s\n", argv[2] );
			return 1;
		}
	}
	else {
		if( argc < 6 ) {
			printUsage();
			return 1;
		}
	}

	/* check bitcount */
	if( argc > 7 ) {
		if( argv[6][0] == '-' ) {
			if( argv[6][1] == 'b' ) {
				sscanf( argv[7], "%d", &b );
				if( ( b == 32 ) || ( b == 24 ) ) {
					bpp = b;
					printf( "set bpp:%d\n", bpp );
				}
				else {
					printf( "unsupported bpp value:%s\n", argv[7] );
					return 1;
				}
			}
		}
	}

	/* open firmware file */
	fpfirm = fopen( argv[1], "rb" );
	if( fpfirm == NULL ) {
		printf( "firmware file open failed.\n" );
		return 1;
	}


	/* batch mode */
	if( mode == 1 ) {
		/* open list file */
		fplist = fopen( argv[3], "rt" );
		if( fplist == NULL ) {
			printf( "list file open failed.\n" );
			return 1;
		}

		/* read list file */
		while( fgets( listarg, sizeof( listarg ), fplist ) != NULL ) {
			if( listarg[ 0 ] != '#' ) {
				/* get parameters */
				sscanf( listarg, "0x%08X 0x%04X %3d %3d %s", &address, &size, &width, &height, &namearg );

				/* read wizpy image data */
				readWizData( fpfirm );

				/* convert to BMP */
				if( bpp == 32 ) {
					/* 32bpp */
					convert32();
				}
				else if( bpp == 24 ) {
					/* 24bpp */
					convert24();
				}
				else {
					printf( "bpp error\n" );
					return 1;
				}

				/* write BMP file */
				sprintf( filename, "%s.bmp", namearg );
				printf("address:0x%08X, width:%d, height:%d, output:%s\n", address, width, height, filename );
				writeBMP( filename );
			}
		}

		/* close list file */
		if( fclose( fplist ) != 0 ) {
			printf( "list file close failed.\n" );
		}
	}
	/* single mode */
	else {
		/* get parameters */
		sscanf( argv[2], "%x", &address );
		sscanf( argv[3], "%d", &width );
		if( width > SCREEN_WIDTH ) {
			printf( "parameter error : width\n" );
		}
		sscanf( argv[4], "%d", &height );
		if( height > SCREEN_HEIGHT ) {
			printf( "parameter error : height\n" );
		}
		printf("address:0x%08X, width:%d, height:%d, output:%s\n", address, width, height, argv[5] );

		/* read wizpy image data */
		readWizData( fpfirm );

		/* convert to BMP */
		if( bpp == 32 ) {
			/* 32bpp */
			convert32();
		}
		else if( bpp == 24 ) {
			/* 24bpp */
			convert24();
		}
		else {
			printf( "bpp error\n" );
			return 1;
		}

		/* write BMP file */
		writeBMP( argv[5] );
	}

	/* close firmware file */
	if( fclose( fpfirm ) != 0 ) {
		printf( "input file close failed.\n" );
	}

	return 0;
}
